/* add recognizer capability to FileWindow  */!!

inherit(FileWindow, #RecWindow, #(recog      /* a Recognizer */
xform      /* a character transform struct */), 2, nil)!!

now(class(RecWindow))!!

now(RecWindow)!!

/* takes a string containing one or more words, creates a
   recognizer object and uses it to build the recog structure */
Def addRecString(self, aStr | wds, tmpRec )
{ wds := words( asUpperCase(aStr) );   /* assume blank delimited words */
  if   size( wds ) = 0
  then errorBox("addRecString","string empty"); ^nil;
  endIf;
  tmpRec := init( new(Recognizer));
  do(wds, {using(w) add(tmpRec, w ) });
  recogBuild( tmpRec );
  recog := cellToStruct( tmpRec );
  ^tmpRec ;
}
!!

/* Redraw the workText from topLine down, and from
  leftChar rightwards, preserve xPos and yPos.  If
  window has focus, show selected text.
  Special version for RecWindow to show hit words */
Def paint(self, hdc | aStr, xH, yH, inst, tmp, p )
{ if   recog
  else ^paint(self:ancestor, hdc );
     /* if no recog defined, use ancestor method */
  endIf;
  xH := xPos;
  yH := yPos;
  home(self);
  initTextColors(self, hdc);
  do(over(topLine, min(size(workText), 1 + topLine + visLines(self))),
  {using(idx)
    inst := chkWorkLine(self, idx, leftChar );
    if   inst
    then tmp := 0 ;
     loop
     while tmp < size(inst)
     begin p := inst[tmp] ;
       if   y(p) > x(p)
       then aStr := copyFrom(workText[idx], x(p), y(p) ) ;
         Call TextOut(hdc, x(self), y(self),aStr, size(aStr) );
         xPos := xPos + size(aStr) ;
       endIf;
        p := inst[ tmp := inc(tmp)] ;  /* now for colored word */ 
       if   y(p) > x(p)   
       then Call SetTextColor(hdc, 0x000000FFL ) ; 
    /* 0x0FFL = red, 0x0FF00L = green, 0x0FF0000L = blue */  
         aStr := copyFrom(workText[idx], x(p), y(p) ) ;
         Call TextOut(hdc, x(self), y(self),aStr, size(aStr) );
         Call SetTextColor(hdc, 0x0L) ;
         xPos := xPos + size(aStr) ;
       endIf;       
       tmp := inc( tmp );
     endLoop; 
            /* finish line which may have punctuation at end */
     if  xPos + leftChar < size(workText[idx]) 
     then aStr := copyFrom(workText[idx], xPos + leftChar,
                              size(workText[idx]) ) ;
     else aStr := " " ;
     endIf;
     Call TextOut(hdc, x(self), y(self),aStr, size(aStr) );
    else
      aStr := copyFrom(workText[idx], leftChar, size(workText[idx]));
      Call TextOut(hdc, x(self), y(self), aStr, size(aStr));
    endIf;
    xPos := 0;
    yPos := yPos + 1;
  });
  xPos := xH;
  yPos := yH;
  setScrollPos(self);
  setHScrollPos(self);
  if hasFocus(self) and isSelText(self)
  then dragDC := hdc;
      invSelTxt(self);
      hideCaret(self);
  endif;
}!!

/* check the nth workText line, return nil if no recog words
   else return collection of points representing paint instructions
   The points are in pairs, first describing characters to be painted 
   normally, then characters to be highlighted.  We use point objects
   because thats the simplest way to hold a pair of integers.
   x = start index of string, y = 1 past last char 
   If x = y, no characters are painted - this might occur if a
   recognized word starts a line.
    */
Def chkWorkLine(self, n , left | tmp, stream, wrd, pos, last, tpos, sz,
    dbData )
{ if   size( workText[n] ) > left 
  then
    tmp :=  xform( asciiz(workText[n]) , xform) ;
      /* creates new upper case string with all seps turned to 0 */
    if   lineRecog(tmp, recog) 
    then stream := streamOver(tmp ) ; /* known to have at least one wd */
            /* make stream from converted work line for detailed look */
    else ^nil ;
    endIf;   
     tmp := new(OrderedCollection, 20 );
     pos := last := 0 ;  /* starting position */
    loop
    while wrd := word(stream, asChar(0))
    begin 
       tpos := position(stream) ;
       if   wdRecog( asciiz(wrd), recog ) 
       then sz := size(wrd)+1;
         if  tpos > left + 1   /* tpos = sep after recog */ 
         then 
          add(tmp, point( max(left, pos ), tpos - sz ) ); 
             /* start@end of normal char before */
          add(tmp, point( max(left, tpos - sz), tpos )) ;
          pos := tpos ;     /* new possible normal */
         endIf;  
       else  last := tpos ; /* track normal words */
       endIf;       
    endLoop;
    if  size(tmp) > 0
    then 
       if   tpos > pos cand tpos > left + 1 
       then  /* finish out line - possibly punctuation, etc. */
          add( tmp,point( max(left,pos), size( workText[n] ) ) );
          add( tmp,point(tpos,tpos) ) ; /* won't display */
       endIf;
         ^tmp ;
    else ^nil ;
    endIf;
  else ^ nil ;  
  endIf;
}!!

/* Draw the given line, preserve xPos and yPos.  Obtain
  own display context.  Note we have to draw an additional space
  in case user deleted a char from end of line.   */
Def drawLine(self, line | xH, yH, aStr, inst, tmp, p, hdc)
{ if recog
  else ^drawLine(self:ancestor, line);
  endIf;
  if line < size(workText)
  then hideCaret(self);
    xH := xPos;
    yH := yPos;
    xPos := 0;
    yPos := line - topLine;
    if   inst := chkWorkLine(self, line, leftChar )
    then tmp := 0 ;
         hdc := getContext(self);
         initTextColors(self, hdc);
     loop
     while tmp < size(inst)
     begin p := inst[tmp] ;
       if   y(p) > x(p)
       then aStr := copyFrom(workText[line], x(p), y(p) ) ;
         Call TextOut(hdc, x(self), y(self),aStr, size(aStr) );
         xPos := xPos + size(aStr) ;
       endIf;
        p := inst[ tmp := inc(tmp)] ;  /* now for colored word */ 
       if   y(p) > x(p)   
       then Call SetTextColor(hdc, 0x000000FFL ) ;
        aStr := copyFrom(workText[line], x(p), y(p) ) ;
        Call TextOut(hdc, x(self), y(self),aStr, size(aStr) );
        Call SetTextColor(hdc, 0x0L) ;
        xPos := xPos + size(aStr) ;
       endIf; 
       tmp := inc( tmp );
     endLoop; 
     /* finish line which may have punctuation or spaces at end */
     if  xPos + leftChar < size(workText[line]) 
     then aStr := copyFrom(workText[line], xPos + leftChar,
              size(workText[line]) ) + " " ;
     else aStr := " "
     endIf;
     Call TextOut(hdc, x(self),y(self),aStr, size(aStr));
     releaseContext(self, hdc);      
    else
     aStr := copyFrom(workText[line], leftChar,
        size(workText[line])) + " ";
     drawText(self, aStr, x(self), y(self)); 
    endIf;
    xPos := xH;
    yPos := yH;
    showCaret(self);
  endif;
}!!

/* initialize the xform struct to convert all lower case to
    upper case letters and major punctuation to 0 */
Def init(self)
{ init(self:ancestor);
  xform := new(Struct,256) ; /* for xform method - initial all 0 */
  do( over(asInt(' '), 129),
     {using(i) putByte(xform,i,i)}) ;  /* printing chars */
  do( over(asInt('a'), asInt('z') + 1 ),
     {using(i) putByte(xform,i-32,i)});  /* lower case to upper */
   /* now for punctuation as separators */  
  do( " !()*+,.:;?\'",
     {using(i) putByte(xform,0,asInt(i)) }); /* punc to null */
  putByte(xform,0, asInt('"') );   /* double quote also punc */
}
!!